/* 
 * UcxSockChan.c
 *
 *  UCX driver for channels based on TCP sockets.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 * Copyright (c) 1997 Computerized Processes Unlimited, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Include required headers

 */
 
#include "tclInt.h"     /* Internal definitions for Tcl. */
#include "tclPort.h"    /* Portability features for Tcl. */

#include <stdio.h>
#include <starlet.h>
#include <descrip.h>
#include <iodef.h>
#include <ucx$inetdef.h>
#include <socket.h>
#include <errno.h>
#include <libdef.h>

#include <ssdef>
#include <stdlib>
#include <stsdef>

#define htons(x) ((unsigned short) ((x<<8)|(x>>8)))


/*
 * Define local macros used
 */

#define htons(x) ((unsigned short) ((x<<8)|(x>>8)))



/*
 * Define local data structures used
 */
typedef struct mysockaddr {
    short   inet_family;
    short   inet_port;
    int     addr;
    char    blkb[8];
} mysockaddr;

typedef struct ItemList {
   int  length;
   struct mysockaddr *host;
} ItemList;

typedef struct ItemList1 {
   int  length;
   char *remote_host;
   int  *return_length;
} ItemList1;

/*
 * This structure describes per-instance state of a tcp based channel.
 */

typedef struct TcpState {
    int flags;              /* ORed combination of the
                                         * bitfields defined below. */
    Tcl_File sock;          /* The socket itself. */
    Tcl_TcpAcceptProc *acceptProc;  /* Proc to call on accept. */
    ClientData acceptProcData;      /* The data for the accept proc. */

    int chan;
    int chanType;
    int checkMask;
    int watchMask;
    int efn;
    short iosbIn[4];
    short iosbOut[4];
    short iosbPeek[4];
    char peekBuff;
    struct sockaddr_in local_host;
    struct sockaddr_in remote_host;
    struct ItemList local_address;
    struct ItemList1 remote_address;
    struct ItemList connect_address;
    int remote_hostaddr[18];
    int retlen;
    void *ptr;  /* Pointer to pending is listen */
} TcpState;

/*
 * These bits may be ORed together into the "flags" field of a TcpState
 * structure.
 */

#define TCP_ASYNC_SOCKET    (1<<0)
#define TCP_ASYNC_CONNECT   (1<<1)
#define TCP_WATCHING_SOCKET (1<<2)
#define TCP_PEEK_PENDING    (1<<3)

/*
 * Only one of the followingmay be set in the chanType field of a TcpState
 * structure.
 */
#define TCP_LISTEN_SOCKET   1
#define TCP_SERVER_CONNECT  2
#define TCP_CLIENT_SOCKET   3

/*
 * Not sure what these are for!
 */
#define SELECT_DONTWAIT     (1<<0)
#define SELECT_READABLE     (1<<1)
#define SELECT_WRITEABLE    (1<<2)
#define SELECT_EXCEPTION    (1<<3)

/*
 * The preferred buffer size for Unix channels.
 */
#define SOCKET_BUFSIZE  4096


/*
 * Static routines for this file:
 */

static TcpState *   CreateSocket _ANSI_ARGS_((Tcl_Interp *interp,
                int port, char *host, int server,
                char *myaddr, int myport));
static int      CreateSocketAddress _ANSI_ARGS_(
                (struct sockaddr_in *sockaddrPtr,
                char *host, int port));
static void     TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int      TcpBlockMode _ANSI_ARGS_((ClientData data,
                    Tcl_File inFile, Tcl_File outFile, int mode));
static int      TcpClose _ANSI_ARGS_((ClientData instanceData,
                Tcl_Interp *interp, Tcl_File inFile,
                Tcl_File outFile));
static int      TcpInput _ANSI_ARGS_((ClientData instanceData,
                    Tcl_File infile, char *buf, int toRead,
                    int *errorCode));
static int      TcpOutput _ANSI_ARGS_((ClientData instanceData,
                    Tcl_File outFile, char *buf, int toWrite,
                    int *errorCode));

static void     socketReadyHandler(TcpState *);
static void     socketFreeProc(ClientData);


/*
 * Define static data
 */
 
/*
 * This structure describes the channel type structure for TCP socket
 * based IO:
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",              /* Type name. */
    TcpBlockMode,           /* Set blocking/nonblocking mode.*/
    TcpClose,               /* Close proc. */
    TcpInput,               /* Input proc. */
    TcpOutput,              /* Output proc. */
    NULL,               /* Seek proc. */
    NULL,               /* Set option proc. */
    NULL                /* Get option proc. */
};

/*
 * The descriptor for the prototype UCX device
 */
struct dsc$descriptor dev = { 3, DSC$K_CLASS_S, DSC$K_DTYPE_T, "BG:" };

short socket_param[2] = { INET$C_TCP, INET_PROTYP$C_STREAM };


/*
 *----------------------------------------------------------------------
 *
 * TcpBlockMode --
 *
 *  Set blocking or non-blocking mode on a TCP channel.
 *
 * Results:
 *  0 if successful, errno when failed.
 *
 * Side effects:
 *  Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static int
TcpBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;        /* Socket state. */
    Tcl_File inFile, outFile;       /* Input, output for channel. */
    int mode;               /* The mode to set. */
{
    TcpState *statePtr;

#ifdef DEBUG
    printf("Entering TcpBlockMode\n");
#endif
    statePtr = (TcpState *) instanceData;
    if (mode == TCL_MODE_BLOCKING) {
        statePtr->flags &= (~(TCP_ASYNC_SOCKET));
    } else {
        statePtr->flags |= TCP_ASYNC_SOCKET;
    }

#ifdef DEBUG
    printf("Leaving TcpBlockMode with %d\n", 0);
#endif
    return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * TcpClose --
 *
 *  Close the socket.
 *
 * Results:
 *  0 if successful, the value of errno if failed.
 *
 * Side effects:
 *  Closes the socket.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static int
TcpClose(instanceData, interp, inFile, outFile)
    ClientData instanceData;    /* The socket to close. */
    Tcl_Interp *interp;     /* For error reporting - unused. */
    Tcl_File inFile, outFile;   /* Unused. */
{
    TcpState *statePtr;
    Tcl_File sockFile;
/*    int sock; */
    int errorCode = 0;
    int status;

    statePtr = (TcpState *) instanceData;
    sockFile = statePtr->sock;
/*    sock = (int) Tcl_GetFileInfo(sockFile, NULL); */
#ifdef DEBUG
  printf("Entering close for %d\n",sockFile);
#endif
    
    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(sockFile);

    /*
     * We assume that inFile==outFile==sockFile and so
     * we only clean up sockFile.
     */

    Tcl_FreeFile(sockFile);

    switch (statePtr->chanType) {
        case TCP_LISTEN_SOCKET: {
        TcpState    *ptr;

        ptr = (TcpState *)statePtr->ptr;
            sys$dassgn (statePtr->chan);
            sys$dassgn (ptr->chan);
            ckfree((char *) ptr);            
            break;      
    }
        case TCP_SERVER_CONNECT: {
            status = sys$qiow (
                        0, 
                        statePtr->chan, 
                        IO$_DEACCESS, 
                        statePtr->iosbIn, 
                        0,0,0,0,0,
                        IO$M_SHUTDOWN, 
                        0,0);
            sys$dassgn (statePtr->chan);
            break;      
    }
        case TCP_CLIENT_SOCKET: {
            status = sys$qiow (
                        0, 
                        statePtr->chan, 
                        IO$_DEACCESS, 
                        statePtr->iosbIn, 
                        0,0,0,0,0,
                        IO$M_SHUTDOWN, 
                        0,0);
            sys$dassgn (statePtr->chan);
            break;      
    }
    }

    ckfree((char *) statePtr);            

#ifdef DEBUG
    printf("Leaving TcpClose with %d.\n", errorCode); 
#endif
    return errorCode;
}


/*
 *----------------------------------------------------------------------
 *
 * TcpInput --
 *
 *  Reads input from the TCP IO channel into the buffer given. Returns
 *  count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *  A count of how many bytes were read is returned and an error
 *  indication is returned in an output argument.
 *
 * Side effects:
 *  Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static int
TcpInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;        /* Socket state. */
    Tcl_File inFile;            /* Input device for channel. */
    char *buf;              /* Where to store data read. */
    int bufSize;            /* How much space is available
                                         * in the buffer? */
    int *errorCode;         /* Where to store error code. */
{
    TcpState *statePtr;         /* The state of the socket. */
    int bytesRead;          /* How many bytes were read? */
    int state;              /* Of waiting for connection. */
    int timeOut;            /* Of waiting for connection. */
    int status;             /* Of qio call */
    int func = IO$_READVBLK;
/*    int sock;*/

#ifdef DEBUG
    printf("Entering TcpInput\n");     
#endif
    *errorCode = 0;
    statePtr = (TcpState *) instanceData;
/*     sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL); */
    
    if (bufSize == 0) {
#ifdef DEBUG
        printf("Leaving TcpInput with %d\n",0);     
#endif
        return 0;
    }

    /*
     * If an asynchronous connect is in progress, attempt to wait for it
     * to complete before reading.
     */
    
    if (statePtr->flags & TCP_ASYNC_CONNECT) {
        if (statePtr->flags & TCP_ASYNC_SOCKET) {
            timeOut = 0;
        } else {
            timeOut = -1;
        }
        state = TclWaitForFile(inFile, TCL_WRITABLE, timeOut);
        if (state & TCL_WRITABLE) {
            statePtr->flags &= (~(TCP_ASYNC_CONNECT));
        } else if (timeOut == 0) {
            *errorCode = EWOULDBLOCK;
#ifdef DEBUG
            printf("Would block 1, leaving TcpInput with %d\n",-1);     
#endif
            return -1;
        }
    }
            
    if (statePtr->flags & TCP_ASYNC_SOCKET) {
#ifdef DEBUG
            printf("Seting nowait on read\n");     
#endif
        func |= IO$M_NOWAIT;
    } else {
#ifdef DEBUG
            printf("Waiting on read of %d\n",bufSize);     
#endif
    }
    status = sys$qiow (
             0, 
             statePtr->chan, 
             func, 
             statePtr->iosbIn, 
             0,0, 
             buf, 
             bufSize, 
             0,0,0,0);
    if ((status & 1) && (statePtr->iosbIn[0] & 1)) {
#ifdef DEBUG
        printf("Leaving 1 TcpInput with %d\n",statePtr->iosbIn[1]);     
#endif
        return statePtr->iosbIn[1];
    } else {
        if (statePtr->iosbIn[0] == SS$_LINKDISCON) {
            /* Convert to EOF */
#ifdef DEBUG
        printf("Leaving TcpInput with EOF\n");     
#endif
            return 0;
        }
        *errorCode = statePtr->iosbIn[1];
#ifdef DEBUG
        printf("Leaving 2 TcpInput with error %d,%d,%d\n",
		status,statePtr->iosbIn[0] ,statePtr->iosbIn[1]);     
#endif
        return -1;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * TcpOutput--
 *
 *  Writes the given output on the TCP IO channel. Returns count of how
 *  many characters were actually written, and an error indication.
 *
 * Results:
 *  A count of how many characters were written is returned and an
 *  error indication is returned in an output argument.
 *
 * Side effects:
 *  Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
TcpOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;        /* Socket state. */
    Tcl_File outFile;           /* Output device for channel. */
    char *buf;              /* The data buffer. */
    int toWrite;            /* How many bytes to write? */
    int *errorCode;         /* Where to store error code. */
{
    TcpState *statePtr;
    int written;
/*    int sock;*/
    int state;              /* Of waiting for connection. */
    int timeOut;            /* Of waiting for connection. */
    int status;             /* Of qio call */
    int func = IO$_WRITEVBLK;
    int socket_errno;

#ifdef DEBUG
    printf("Entering TcpOutput\n");     
#endif
    *errorCode = 0;
    socket_errno = 0;
    statePtr = (TcpState *) instanceData;
/*    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL); */

    /*
     * If an asynchronous connect is in progress, attempt to wait for it
     * to complete before reading.
     */
    
    if (statePtr->flags & TCP_ASYNC_CONNECT) {
        if (statePtr->flags & TCP_ASYNC_SOCKET) {
            timeOut = 0;
        } else {
            timeOut = -1;
        }
        state = TclWaitForFile(outFile, TCL_WRITABLE, timeOut);
        if (state & TCL_WRITABLE) {
            statePtr->flags &= (~(TCP_ASYNC_CONNECT));
        } else if (timeOut == 0) {
            *errorCode = socket_errno = EWOULDBLOCK;
#ifdef DEBUG
            printf("Leaving 0 TcpOutput with %d\n",-1);     
#endif
            return -1;
        }
    }

    if (statePtr->flags & TCP_ASYNC_SOCKET) {
        func |= IO$M_NOWAIT;
    }
    status = sys$qiow (
             0, 
             statePtr->chan, 
             func, 
             statePtr->iosbIn, 
             0,0, 
             buf, 
             toWrite, 
             0,0,0,0);
    if ((status & 1) && (statePtr->iosbIn[0] & 1)) {
#ifdef DEBUG
        printf("Leaving 1 TcpOutput with %d\n",statePtr->iosbIn[1]);     
#endif
        return statePtr->iosbIn[1];
    } else {
        *errorCode = statePtr->iosbIn[1];
#ifdef DEBUG
        printf("Leaving 2 TcpOutput with %d\n",-1);     
#endif
        return -1;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * CreateSocket --
 *
 *  This function opens a new socket in client or server mode
 *  and initializes the TcpState structure.
 *
 * Results:
 *  Returns a new TcpState, or NULL with an error interp.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------
 */

static TcpState *
CreateSocket(interp, port, host, server, myaddr, myport)
    Tcl_Interp *interp; /* For error reporting; can be NULL. */
    int port;           /* Port number to open. */
    char *host;         /* Name of host on which to open port. */
    int server;         /* 1 if socket should be a server socket,
                 * else 0 for a client socket. */
    char *myaddr;       /* Optional client-side address */
    int myport;         /* Optional client-side port */
{
    int enable = 1;
    int socketValid = 1;
    int status, chan, maxAccept, reuse, curState, origState;
    struct sockaddr_in sockaddr; /* Socket address */
    TcpState *statePtr;
    TcpState *temp;

#ifdef DEBUG
    printf("Enterning CreateSocket\n");     
#endif

    /*
     * Assign a device to for the socket
     */
    status = sys$assign (
              &dev, 
              &chan, 
              0, 
              0);
        socketValid = status & 1;
#ifdef DEBUG
        printf ("Assign: status = %d, chan = %d, boo=%d\n",
        status, chan, socketValid);  
#endif
     
    if (socketValid) {
        /*
         * Allocate a new TcpState for this socket.
         */

        statePtr = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
	statePtr->efn = getEventFlag();
        statePtr->flags = 0;
        statePtr->sock = Tcl_GetFile((ClientData) chan, TCL_VMS_SOCKET);
        statePtr->chan = chan;
        statePtr->checkMask = 0;
        statePtr->watchMask = 0;
        statePtr->local_address.length = sizeof(statePtr->local_host);
        statePtr->local_address.host = &(statePtr->local_host);
        Tcl_SetNotifierData(statePtr->sock, socketFreeProc, (ClientData)statePtr);
    }
    
    if (server) {
        maxAccept = 99;
        reuse = 0;
        myport = port;
        statePtr->chanType = TCP_LISTEN_SOCKET;
    } else {
        reuse = SOCKOPT$M_REUSEADDR;
        maxAccept = 0;
    }
    

    if (socketValid && !server) {
        if (! CreateSocketAddress(&statePtr->remote_host, host, port)) {
#ifdef DEBUG
            printf ("Create Socket Address for remote failed!\n"); 
#endif
            socketValid = 0;
        }
    }
    
    if (socketValid) {
        if ((myaddr != NULL || myport != 0) &&
                ! CreateSocketAddress(&statePtr->local_host, myaddr, myport)) {
#ifdef DEBUG
            printf ("Create Socket Address for local failed!\n"); 
#endif
            socketValid = 0;
        }
    }

    if (socketValid) {
        status = sys$qiow (
                  0, 
                  chan, 
                  IO$_SETMODE, 
                  statePtr->iosbIn, 
                  0, 
                  0, 
                  &socket_param, 
                  reuse, 
                  &statePtr->local_address, 
                  maxAccept, 
                  0, 
                  0);
#ifdef DEBUG
        printf ("Set Mode: chan =  %d, status = %d, iosb = %d,%d\n",
        chan, status, statePtr->iosbIn[0], statePtr->iosbIn[1]);
#endif
        socketValid = (status & 1) & (statePtr->iosbIn[0] & 1);
    }
    if (socketValid) {
        if (server) {
            temp = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
	    temp->efn = getEventFlag();
            temp->remote_address.length = 16;
            temp->remote_address.remote_host = &((char ) temp->remote_hostaddr[0]);
            temp->remote_address.return_length = &(temp->retlen);
            temp->flags = 0;
            temp->checkMask = 0;
            temp->watchMask = 0;
            statePtr->ptr = (void *)temp;
            status = sys$assign (
                      &dev, 
                      &temp->chan, 
                      0, 
                      0);
#ifdef DEBUG
            printf ("Assign child: status = %d, chan=%d\n", status, temp->chan); 
#endif
            status = sys$qio (
                      statePtr->efn, 
                      statePtr->chan, 
                      IO$_ACCESS|IO$M_ACCEPT, 
                      statePtr->iosbIn, 
                      0, 
                      0, 
                      0, 
                      0,
                      &(temp->remote_address), 
                      &(temp->chan), 
                      0, 
                      0);
#ifdef DEBUG
            printf ("Server Access: chan =  %d, status = %d, iosb = %d,%d\n",
                    chan, status, statePtr->iosbIn[0], statePtr->iosbIn[1]);
#endif
            socketValid = (status & 1);
        } else {
            statePtr->connect_address.length = sizeof(statePtr->remote_host);
            statePtr->connect_address.host = &(statePtr->remote_host);
            status = sys$qiow (
                      0,
                      chan, 
                      IO$_ACCESS, 
                      statePtr->iosbIn, 
                      0, 
                      0, 
                      0, 
                      0, 
                      &statePtr->connect_address, 
                      0, 
                      0, 
                      0);
#ifdef DEBUG
                printf ("Client Access: chan =  %d, status = %d, iosb = %d,%d\n",
                chan, status, statePtr->iosbIn[0], statePtr->iosbIn[1]);
#endif
                socketValid = (status & 1) & (statePtr->iosbIn[0] & 1);
        }
    }

    /*
     * If we could not create a valid socket
     */
#ifdef DEBUG
    printf ("At test boo=%d\n", socketValid); 
#endif
    if (!socketValid) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                Tcl_PosixError(interp), (char *) NULL);
        }
#ifdef DEBUG
        printf("Leaving CreateSocket with bad data\n");     
#endif
        return NULL;
    }
   
#ifdef DEBUG
    printf("Leaving CreateSocket\n");     
#endif
    return statePtr;

}


/*
 *----------------------------------------------------------------------
 *
 * CreateSocketAddress --
 *
 *  This function initializes a sockaddr structure for a host and port.
 *
 * Results:
 *  1 if the host was valid, 0 if the host could not be converted to
 *  an IP address.
 *
 * Side effects:
 *  Fills in the *sockaddrPtr structure.
 *
 *----------------------------------------------------------------------
 */

static int
CreateSocketAddress(sockaddrPtr, host, port)
    struct sockaddr_in *sockaddrPtr; /* Socket address */
    char *host;             /* Host.  NULL implies INADDR_ANY */
    int port;               /* Port number */
{
    struct hostent *hostent;        /* Host database entry */
    struct in_addr addr;        /* For 64/32 bit madness */

#ifdef DEBUG
    printf("Entering CreateSocketAddress\n"); 
#endif
    (void) memset((char *) sockaddrPtr, '\0', sizeof(struct sockaddr_in));
    sockaddrPtr->sin_family = AF_INET;
    sockaddrPtr->sin_port = htons((unsigned short) (port & 0xFFFF));
    if ((host == NULL) || (strcmp(host,"") == 0)) {
        addr.s_addr = INADDR_ANY;
    } else {
        hostent = gethostbyname(host);
        if (hostent != NULL) {
            memcpy((char *) &addr,
                    (char *) hostent->h_addr_list[0], 
                    (size_t) hostent->h_length);
        } else {
            addr.s_addr = inet_addr(host);
            if (addr.s_addr == (unsigned long)-1) {
#ifdef DEBUG
                printf ("Bad address (-1)\n"); 
#endif
                return 0;   /* error */
            }
        }
    }

    sockaddrPtr->sin_addr.s_addr = addr.s_addr;
#ifdef DEBUG
    printf("Leaving CreateSocketAddress with %d\n",1); 
#endif
    return 1;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *  Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *  The channel or NULL if failed.  An error message is returned
 *  in the interpreter on failure.
 *
 * Side effects:
 *  Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host, myaddr, myport, async)
    Tcl_Interp *interp;         /* For error reporting; can be NULL. */
    int port;               /* Port number to open. */
    char *host;             /* Host on which to open port. */
    char *myaddr;           /* Client-side address */
    int myport;             /* Client-side port */
    int async;		    /* Flag indicating if asynchronus operation */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */
#ifdef DEBUG
    printf("Enterning Tcl_OpenTcpClient\n");     
#endif
    statePtr = CreateSocket(interp, port, host, 0, myaddr, myport);
    if (statePtr == NULL) {
        return NULL;
    }

    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;

    sprintf(channelName, "sock%d",
        (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, statePtr->sock,
        statePtr->sock, (ClientData) statePtr);
/*
    if (chan != (Tcl_Channel) NULL) {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }
*/

#ifdef DEBUG
    printf("Leavinging Tcl_OpenTcpClient\n");     
#endif
    return chan;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *  Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *  The channel or NULL if failed.
 *
 * Side effects:
 *  Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, host, acceptProc, acceptProcData)
    Tcl_Interp *interp;         /* For error reporting - may be
                                         * NULL. */
    int port;               /* Port number to open. */
    char *host;             /* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;  /* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;      /* Data for the callback. */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

#ifdef DEBUG
    printf("Enterning Tcl_OpenTcpServer\n");     
#endif
    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, host, 1, NULL, 0);
    if (statePtr == NULL) {
    return NULL;
    }

    statePtr->acceptProc = acceptProc;
    statePtr->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients. The caller will use Tcl_TcpRegisterCallback
     * to register a callback to call when a new connection is
     * accepted.
     */

    Tcl_CreateFileHandler(statePtr->sock, TCL_READABLE, TcpAccept,
            (ClientData) statePtr);
    
    sprintf(channelName, "sock%d",
        (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) statePtr);
/*
    if (chan != (Tcl_Channel) NULL) {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }
*/

#ifdef DEBUG
    printf("Leaving Tcl_OpenTcpServer for %d\n",chan);     
#endif
    return chan;
}


/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *  Accept a TCP socket connection.  This is called by the event loop,
 *  and it in turns calls any registered callbacks for this channel.
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  Evals the Tcl script associated with the server socket.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;            /* Callback token. */
    int mask;               /* Not used. */
{
    TcpState *LstatePtr;        /* Client data of server socket. */
    TcpState *temp;        /* Client data of server socket. */
    TcpState *NstatePtr;        /* Client data of server socket. */
    int newsock;            /* The new client socket */
    TcpState *newSockState;     /* State for new socket. */
    struct sockaddr_in addr;        /* The remote address */
    int len;                /* For accept interface */
    Tcl_Channel chan;           /* Channel instance created. */
    char channelName[20];

    LstatePtr = (TcpState *) data;  /* Get Listen Info */
#ifdef DEBUG
    printf("Entering TclAccpet for %d\n",LstatePtr->chan);     
#endif
    NstatePtr = (TcpState *)LstatePtr->ptr;      /* Get Info on Accepted Channel */
    NstatePtr->chanType = TCP_SERVER_CONNECT;

    /*
     * Set up to accept another connection
     */
    temp = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    temp->efn = getEventFlag();
    temp->remote_address.length = 16;
    temp->remote_address.remote_host = &((char) temp->remote_hostaddr[0]);
    temp->remote_address.return_length = &(temp->retlen);
    temp->flags = 0;
    temp->checkMask = 0;
    temp->watchMask = 0;
    sys$assign (
              &dev, 
              &temp->chan, 
              0, 
              0);
#ifdef DEBUG
    printf ("Assign child: chan=%d\n", temp->chan); 
#endif
    LstatePtr->ptr = (void *)temp;
    sys$qio (
              LstatePtr->efn, 
              LstatePtr->chan, 
              IO$_ACCESS|IO$M_ACCEPT, 
              LstatePtr->iosbIn, 
              0, 
              0, 
              0, 
              0,
              &(temp->remote_address), 
              &(temp->chan), 
              0, 
              0);
#ifdef DEBUG
    printf ("Server Access: chan =  %d, iosb = %d,%d\n",
            temp->chan, LstatePtr->iosbIn[0], LstatePtr->iosbIn[1]);
#endif

    /*
     * Process connection just accepted
     */
    NstatePtr->sock = Tcl_GetFile((ClientData) NstatePtr->chan, TCL_VMS_SOCKET);
    sprintf(channelName, "sock%d", (int) NstatePtr->sock);
    Tcl_SetNotifierData(NstatePtr->sock, socketFreeProc, (ClientData)NstatePtr);
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NstatePtr->sock,
                NstatePtr->sock, (ClientData) NstatePtr);
#ifdef DEBUG
    printf("Accepted '%s' for chan %d\n",channelName,LstatePtr->chan);     
#endif
/*
    Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
*/
    if (LstatePtr->acceptProc != (Tcl_TcpAcceptProc *) NULL) {
        NstatePtr->remote_host.sin_port = NstatePtr->remote_hostaddr[0];
        NstatePtr->remote_host.sin_addr.S_un.S_addr = NstatePtr->remote_hostaddr[1];
        (LstatePtr->acceptProc) (LstatePtr->acceptProcData, 
                     chan,
                     inet_ntoa(NstatePtr->remote_host.sin_addr), 
                     ntohs(NstatePtr->remote_host.sin_port)) ;
    }
#ifdef DEBUG
    printf("Leaving TclAccept for chan %d\n",LstatePtr->chan);     
#endif
}


static void
socketFreeProc(ClientData clientdata)
{
    /*
     * Free up socket info structure
     */
    TcpState *statePtr = (TcpState *)clientdata;

    if (statePtr) {
       freeEventFlag(statePtr->efn);
       ckfree((char *)statePtr);
    }
#ifdef DEBUG
    printf("Leaving socketFreeProc\n");     
#endif
}


TclVmsWatchSocket(Tcl_File file, int mask)
{
    /*
     * Watch for something to happen on a socket.
     *
     * Start a QIO and return the event flag
     * that will be set when something happens.
     */
    int status;
    
    TcpState *statePtr = (TcpState *)Tcl_GetNotifierData(file, NULL);
#ifdef DEBUG
    printf("Entering TclVmsWatchSocket for %d\n",statePtr->chan);     
#endif
    status = 1;
    
    statePtr->watchMask = 0;
    if (mask & TCL_READABLE) {
        statePtr->watchMask |= SELECT_READABLE;
        if ((statePtr->chanType != TCP_LISTEN_SOCKET) &
	    !(statePtr->flags & TCP_PEEK_PENDING)
  	   ) {
	    statePtr->flags |= TCP_PEEK_PENDING;
            status = sys$qio(statePtr->efn,
                             statePtr->chan, 
                             IO$_READVBLK,
                             statePtr->iosbPeek,
                             socketReadyHandler,
                             statePtr,
                             &(statePtr->peekBuff),
                             1, 
                             0, 
                             UCX$M_MSG_PEEK,
                             0, 0);
	}
    }
    if (mask & TCL_WRITABLE)
    statePtr->watchMask |= SELECT_WRITEABLE;
    if (mask & TCL_EXCEPTION)
    statePtr->watchMask |= SELECT_EXCEPTION;
    if (status & 1)
        statePtr->flags |= TCP_WATCHING_SOCKET;
    else
        printf("IO$SELECT failed: %x\n", status); 
#ifdef DEBUG
    printf("Leaving TclVmsWatchSocket for %d\n",statePtr->chan);     
#endif
    return statePtr->efn;
}


int VmsSocketReady(Tcl_File file, int mask)
{
    /*
     * Check and see what's ready for a socket
     */
    int readyMask = 0;
    TcpState *statePtr = (TcpState *)Tcl_GetNotifierData(file, NULL);
    sys$clref(statePtr->efn);
#ifdef DEBUG
    printf("Entering VmsSocketReady for %d\n",statePtr->chan);     
#endif

    /*
     *Convert local flags to TCL flags
     */
    if ((statePtr->watchMask & SELECT_READABLE)){
        if (statePtr->chanType != TCP_LISTEN_SOCKET) {
            if (statePtr->iosbPeek[0] != 0) {
		statePtr->flags &= !TCP_PEEK_PENDING;
                readyMask |= TCL_READABLE;
		statePtr->iosbPeek[0] = 0;
            }
        } else {
            if (statePtr->iosbIn[0] != 0) {
                readyMask |= TCL_READABLE;
            }
        } 
    }
    if (statePtr->watchMask & SELECT_WRITEABLE)
        readyMask |= TCL_WRITABLE;
    if (statePtr->watchMask & SELECT_EXCEPTION)
        readyMask |= TCL_EXCEPTION;
#ifdef DEBUG
    printf("Leaving VmsSocketReady for %d\n",statePtr->chan);     
#endif
    return (readyMask & mask);
}


static void
socketReadyHandler(TcpState *statePtr)
{
#ifdef DEBUG
    printf("In socketReadyHandler for %d\n",statePtr->chan);     
#endif
    /*
     * AST routine called when something is ready on a socket.
     */
    if (statePtr->iosbPeek[0] & 1)
        statePtr->flags &= ~TCP_WATCHING_SOCKET;
}
